/*  Copyright 2009 Marc Toussaint
    email: mtoussai@cs.tu-berlin.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a COPYING file of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/> */

#include "functions.h"
#include "algos.h"

double NNsdv(const double& a,const double& b,double sdv){
  double d=(a-b)/sdv;
  double norm = 1./(::sqrt(MT_2PI)*sdv);
  return norm*EXP(-.5*d*d);
}
double NNsdv(double x,double sdv){
  x/=sdv;
  double norm = 1./(::sqrt(MT_2PI)*sdv);
  return norm*EXP(-.5*x*x);
}
double NNinv(const arr& a,const arr& b,const arr& invC){
  arr d=a-b;
  double norm = ::sqrt(determinant((1./MT_2PI)*invC));
  return norm*EXP(-.5*scalarProduct(invC,d,d));
}
double logNNinv(const arr& a,const arr& b,const arr& invC){
  NIY;
  return 1;
  /*
  arr d=a-b;
  double norm = ::sqrt(fabs(MT::determinant_LU((1./MT_2PI)*invC)));
  return ::log(norm) + (-.5*scalarProduct(invC,d,d));
  */
}
double logNNprec(const arr& a,const arr& b,double prec){
  uint n=a.N;
  arr d=a-b;
  double norm = pow(prec/MT_2PI,.5*n);
  return ::log(norm) + (-.5*prec*scalarProduct(d,d));
}
double logNN(const arr& a,const arr& b,const arr& C){
  arr invC;
  inverse_SymPosDef(invC,C);
  return logNNinv(a,b,invC);
}
double NN(const arr& a,const arr& b,const arr& C){
  arr invC;
  inverse_SymPosDef(invC,C);
  return NNinv(a,b,invC);
}
//! non-normalized!! Gaussian function (f(0)=1)
double NNNNinv(const arr& a,const arr& b,const arr& invC){
  arr d=a-b;
  return EXP(-.5*scalarProduct(invC,d,d));
}
double NNNN(const arr& a,const arr& b,const arr& C){
  arr invC;
  inverse_SymPosDef(invC,C);
  return NNNNinv(a,b,invC);
}
double NNzeroinv(const arr& x,const arr& invC){
  double norm = ::sqrt(determinant((1./MT_2PI)*invC));
  return norm*EXP(-.5*scalarProduct(invC,x,x));
}
//! gradient of a Gaussian
double dNNinv(const arr& x,const arr& a,const arr& invA,arr& grad){
  double y=NNinv(x,a,invA);
  grad = y * invA * (a-x);
  return y;
}
//! gradient of a non-normalized Gaussian
double dNNNNinv(const arr& x,const arr& a,const arr& invA,arr& grad){
  double y=NNNNinv(x,a,invA);
  grad = y * invA * (a-x);
  return y;
}
double NNsdv(const arr& a,const arr& b,double sdv){
  double norm = 1./(::sqrt(MT_2PI)*sdv);
  return norm*EXP(-.5*sqrDistance(a,b)/(sdv*sdv));
}
double NNzerosdv(const arr& x,double sdv){
  double norm = 1./(::sqrt(MT_2PI)*sdv);
  return norm*EXP(-.5*sumOfSqr(x)/(sdv*sdv));
}
