/*  Copyright 2009 Marc Toussaint
    email: mtoussai@cs.tu-berlin.de

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a COPYING file of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/> */

#include "soc.h"
#include "ors.h"
#include "algos.h"
#ifdef MT_HSL
#  include "hslModule.h"
#endif

double attractorCostAndGradient(soc::SocAbstraction& soci, arr& dCdvia, arr& q, const ors::Spline& spline,const arr& via,bool costOnly){ NIY; }

//===========================================================================

void optimizeTrajUsingRprop(soc::SocAbstraction& soci,
                            arr& q,
                            uint maxIterations,
                            double stoppingTolerance,
                            uint display){
  
  uint T=soci.nTime(),n=soci.qDim();
  
  MT::timerStart();
  
  //initialize trajectory
  if(q.N){
    CHECK(q.nd==2 && q.d0==T && q.d1==n,"");
  }else{
    q.resize(T,n);
    //soc::bayesianIKTrajectory(soci,q);
    soc::straightTaskTrajectory(soci,q,0);
  }
  if(soci.os){
    *soci.os <<std::setw(3) <<-1 <<"  time "  <<MT::timerRead(false);
    soci.computeTotalCost(q);
  }
  if(soci.gl){
    soci.displayTrajectory(q,display,STRING("gradient optimization -- iteration "<<-1));
  }
  
  uint k;
  arr dCdq;
  Rprop rprop;
  rprop.init(1e-3);
  for(k=0;k<maxIterations;k++){
    soci.computeTotalCostGradient(dCdq,q);
    rprop.step(q,dCdq);
    soci.computeTotalCost(q);
    
    if(soci.os){
      *soci.os <<std::setw(3) <<k <<"  time " <<MT::timerRead(false);
      soci.computeTotalCost(q);
    }
    if(soci.gl){
      soci.displayTrajectory(q,display,STRING("gradient optimization -- iteration "<<k));
    }
  }
}

//===========================================================================

namespace minProblemStatic{
  uint count;
  arr q,dq;
  ostream *os;
  OpenGL *gl;
  ors::Spline *spline;
  uint display;

  double f(const arr &x,void *data){
    ((soc::SocAbstraction*)data)->getq0(x[0]());
    if(!spline) q=x; else q=spline->basis*x;
    if(gl) ((soc::SocAbstraction*)data)->displayTrajectory(q,display,STRING("gradient optimization -- iteration "<<-1));
    if(os) (*os) <<std::setw(3) <<count++ <<"  time " <<MT::timerRead();
    if(os!=&cout) cout <<'.' <<std::flush;
    return ((soc::SocAbstraction*)data)->computeTotalCost(q);
  }

  void df(arr &dx,const arr &x,void *data){
    ((soc::SocAbstraction*)data)->getq0(x[0]());
    if(!spline) q=x; else q=spline->basis*x;
    ((soc::SocAbstraction*)data)->computeTotalCostGradient(dq,q);
    if(!spline) dx=dq; else dx=spline->basis_trans*dq;
    dx[0]()=0.;
  }

  double f_att(const arr &x,void *data){
    arr q0;
    ((soc::SocAbstraction*)data)->getq0(q0);
    ((soc::SocAbstraction*)data)->setq(q0);
    ((soc::SocAbstraction*)data)->getPhi(x[0](),0);
    double C;
    static arr dummy;
    soc::SocAbstraction& soci=*((soc::SocAbstraction*)data);
    if(os) (*os) <<std::setw(3) <<count++ <<"  time " <<MT::timerRead();
    C=attractorCostAndGradient(soci,dummy,q,*spline,x,true);
    if(gl){
      /*plotClear();
      plotPoints(x);
      plotLine(x);*/
      NIY;
      ((soc::SocAbstraction*)data)->displayTrajectory(q,display,STRING("gradient optimization -- iteration "<<-1));
    }
    //return ((soc::SocAbstraction*)data)->computeTotalCost(q);
    return C;
  }

  void df_att(arr &dx,const arr &x,void *data){
    //((soc::SocAbstraction*)data)->getq0(x[0]());
    soc::SocAbstraction& soci=*((soc::SocAbstraction*)data);
    attractorCostAndGradient(soci,dx,q,*spline,x,false);
    dx[0]()=0.;
  }

  void c(arr &cx,const arr &x,void *data){ cx.resize(0); }
  void dc(arr &dcx,const arr &x,void *data){ dcx.resize(0,x.N); }
}

//===========================================================================

//! compute an optimized trajectory using gradient descent
void soc::gradientOptimization(SocAbstraction& soci,
                                arr& q,
                                int gradient_method,
                                uint maxIterations,
                                uint spline_points,
                                uint spline_degree,
                                double stoppingTolerance,
                                bool checkGradient,
                                uint display){

  uint T=soci.nTime(),n=soci.qDim();
  
  MT::timerStart();
  
  //initialize trajectory
  if(q.N){
    CHECK(q.nd==2 && q.d0==T && q.d1==n,"");
  }else{
    q.resize(T,n);
    if(!soci.dynamic){
      soc::bayesianIKTrajectory(soci,q);
      //soc::straightTaskTrajectory(soci,q,0);
    }else{
      soci.getq0(q[0]());
      for(uint t=1;t<T;t++) q[t]()=q[0];
    }
  }
  if(soci.os){
    *soci.os <<std::setw(3) <<-1 <<"  time " <<MT::timerRead(false);
    soci.computeTotalCost(q);
  }
  if(soci.gl){
    soci.displayTrajectory(q,display,STRING("gradient optimization -- iteration "<<-1));
  }
  
  bool task_spline=false;
  if(gradient_method==Attractor) task_spline=true;

  ors::Spline spline;
  if(spline_points){
    uint K=spline_points;
    spline.setUniformNonperiodicBasis(T-1,K,spline_degree);
    if(!task_spline){
      spline.points.resize(K+1,soci.qDim());
      spline.points[0]() = q[0];
      spline.points[K]() = q[T-1];
      for(uint k=1;k<K;k++)
        spline.points[k]() = q[(uint)((double)k*T/K)];
    }else{
      spline.points.resize(K+1,soci.yDim(0));
      soci.setq(q[0]);   soci.getPhi(spline.points[0](),0);
      soci.setq(q[T-1]); soci.getPhi(spline.points[K](),0);
      for(uint k=1;k<K;k++){
        soci.setq(q[(uint)((double)k*T/K)]); soci.getPhi(spline.points[k](),0);
      }
    }
  }

  minProblemStatic::count = 0;
  minProblemStatic::os = soci.os;
  minProblemStatic::gl = soci.gl;
  minProblemStatic::display = display;
  
  arr x; //variable to be optimized
  if(spline_points || gradient_method==Attractor){
    x = spline.points;
    minProblemStatic::spline = &spline;
  }else{
    x = q;
    minProblemStatic::spline = NULL;
  }
  
  double fmin;
  switch(gradient_method){
  case ConjGrad:
  case LevMar:
  case Rprop:
  case RpropConjGrad:
    MT::minimize(minProblemStatic::f,
                 minProblemStatic::df,
                 &soci,
                 x,
                 &fmin,
                 gradient_method,
                 maxIterations,
                 stoppingTolerance,
                 checkGradient);
    break;
  case SQP:
#ifdef MT_HSL
    hsl_SQP_vf13(minProblemStatic::f,
                 minProblemStatic::df,
                 minProblemStatic::c,
                 minProblemStatic::dc,
                 &soci,
                 x,
                 0,0,maxIterations,
                 1e-4);
#else
        NIY;
#endif
  break;
  case Attractor:
    MT::minimize(minProblemStatic::f_att,
                 minProblemStatic::df_att,
                 &soci,
                 x,
                 &fmin,
                 2,
                 maxIterations,
                 stoppingTolerance,
                 checkGradient);
    break;
  default: NIY;
  }

  if(spline_points) q = spline.basis*x;
}

//===========================================================================
