#include <MT/ors.h>
#include <MT/ors_control.h>
#include <MT/ors_opengl.h>
#include <MT/ors_swift.h>
#include <MT/soc.h>
#include <MT/opengl.h>

int main(int argn,char **argv){
  //-- setup the simulator, swift, opengl, and the SocAbstraction
  ors::Graph ors;
  ors.init("test.ors");

  SwiftModule swift;
  swift.init(ors,.5);
  
  OpenGL gl;
  gl.add(glStandardScene);
  gl.add(ors::glDrawGraph,&ors);
  gl.setPosition(5,-10,10);
  gl.focus(0,0,1);
  gl.watch();

  uint T=200;
  soc::OrsSocImplementation soc;
  soc.initKinematic(&ors,&swift,&gl,T);
  soc.os=&std::cout;

  //-- setup the control variables (problem definition)
  ControlVariable *pos = new ControlVariable("position",ors, posCVT,"arm7","<t(0 0 .2)>",0,0,ARR());
  pos->setAttractorDynamics(20,.2);
  pos->x_target = arr(ors.getName("target")->X.p.v,3); //ARR(1.,0.,1.);
  
  ControlVariable *col = new ControlVariable("collision",ors, collCVT,0,0,0,0,ARR(.1));
  col->setGains(.5,.0);
  col->prec=1e6;
  col->x_target = ARR(0.);
  
  soc.setControlVariables(TUPLE(pos,col));
  
  //-- inverse kinematics to reach the targets
  arr q,dq;
  soc.getq0(q);
  for(uint t=0;t<T;t++){
    soc::bayesianIKControl(soc,dq,0);
    q += dq;
    soc.setq(q);
    soc.reportOnState(cout);
    gl.update(STRING("bayesianIK: iteration "<<t));
    //gl.watch();
  }
  gl.watch();
  
  //-- planning (AICO) to generate an optimal (kinematic) trajectory
  soc.getq0(q);
  soc.setq(q);
  pos->setInterpolatedTargetTrajectory(T);
  pos->setPrecisionTrajectoryFinal(T,1e-2,1e4);
  col->setInterpolatedTargetTrajectory(T);
  col->setPrecisionTrajectoryConstant(T,1e6);

  q.clear();
  AICO_kinematic(soc,q,10,.9,.01,0);
  soc.displayTrajectory(q,1,"AICO (planned trajectory)");
  
  return 0;
}
